package xml.xacml;

import org.xml.sax.SAXException;

import javax.xml.parsers.ParserConfigurationException;
import javax.xml.parsers.SAXParser;
import javax.xml.parsers.SAXParserFactory;
import java.io.ByteArrayInputStream;
import java.io.IOException;
import java.io.File;
import java.util.Vector;

import futils.Futil;
import futils.WriterUtil;
import futils.ReaderUtil;

/**
 * Copyright 2005, DocJava, Inc.
 * User: dlyon
 * Date: Nov 15, 2005
 * Time: 3:49:37 PM
 * ACL - access control list. Role is to keep a list of people
 * who are allowed to access the system.
 * Features include:
 * <code>
 * 1. Check if a person is in the list.
 * 2. Add a person to the ACL.
 * 3. Remove a person from the ACL.
 * 4. Save the ACL to a file/string.
 * 5. Make sure uids are unique!
 * 6. Restore the ACL from a file/string
 * <p/>
 * </code>
 */
public class Acl {
    private Vector v = new Vector();

    public void add(Person p) throws DuplicatePersonException {
        if (isUidUnique(p)) {
            v.addElement(p);
            return;
        }
        throw new DuplicatePersonException(p + " already in list");
    }

    public void remove(Person p) throws PersonNotInListException {
        if (! isInList(p)) throw new PersonNotInListException(p + " not in list, can't remove");
        v.remove(p);
    }

    public boolean isUidUnique(Person p) {
        for (int i = 0; i < v.size(); i++) {
            Person piList = (Person) v.elementAt(i);
            if (piList.getUid().equals(p.getUid())) return false;
        }
        return true;
    }

    public Person[] getList() {
        Person p[] = new Person[v.size()];
        v.copyInto(p);
        return p;
    }

    public boolean isInList(Person p) {
        return v.contains(p);
    }
    public void save(){
        File f =Futil.getWriteFile("select a xml output location");
        save(f);
    }

    public void save(File f) {
        WriterUtil.writeString(f,toXACML());
    }

    public String toXACML() {
        StringBuffer sb = new StringBuffer("<?xml version=\"1.0\" encoding=\"ISO-8859-1\" standalone=\"yes\"?>\n" +
                "\n" +
                "<!DOCTYPE acl [\n" +
                "        <!ELEMENT person (uid, password,email)>\n" +
                "        <!ELEMENT uid (#PCDATA)>\n" +
                "        <!ELEMENT password (#PCDATA)>\n" +
                "        <!ELEMENT email (#PCDATA)>\n" +
                "        <!ELEMENT acl (person*)>\n" +
                "        ]>\n" +
                "<acl>\n");
        Person p[] = getList();
        for (int i = 0; i < p.length; i++)
            sb.append(p[i].toXACML());
        sb.append("</acl>\n");
        return sb.toString();
    }

    public void print() {
        Person p[] = getList();
        for (int i = 0; i < p.length; i++)
            System.out.println(p[i]);
    }

    public static void main(String[] args) throws DuplicatePersonException,
            IOException, ParserConfigurationException, SAXException {
        Acl a = new Acl();
        Person p = new Person();
        p.setUid("lyon");
        p.setPw("foo");
        p.setEmail("lyon@docjava.com");
        a.add(p);
        p = new Person();
        p.setUid("foo foo");
        p.setPw("foo");
        p.setEmail("foo@docjava.com");
        a.add(p);

        String xacmlString = a.toXACML();
        //System.out.println(xacmlString);
        Acl a2 = Acl.restore(xacmlString);
        a2.save();
        Acl a3 = Acl.restore();
        a3.print();
        System.out.println("done!");
    }
    public static Acl restore() throws IOException, ParserConfigurationException, SAXException {
        File f = Futil.getReadFile("select an acl location");
        return restore(ReaderUtil.getFileAsOneBigString(f));
    }
    public static Acl restore(String s)
            throws ParserConfigurationException, SAXException, IOException {
        SAXParserFactory spf
                = SAXParserFactory.newInstance();
        spf.setValidating(true);

        SAXParser sp = spf.newSAXParser();

        ACLHandler dh = new ACLHandler();
        byte b[] = s.getBytes();
        ByteArrayInputStream bis = new ByteArrayInputStream(b);

        sp.parse(bis, dh);
        return dh.getAcl();
    }
}
